function [param rslt] = HSIC_CDA(X, Y, varargin)

% HSIC_CDA - CDA by maximizing HSIC
%
% 'dim_U'         Dimension of projection matrix for X      1
% 'dim_V'         Dimension of projection matrix for Y      1
% 'sigma_X'       Gaussian kernel parameters for X
% 'sigma_Y'       Gaussian kernel parameters for Y     
% 'kappa'         Regularization parameters           
% 'trial'         The number of trials                      10
% 'iter'          The maximum number of iteration           200

% --------------------------------------------------

  param.N = size(X,2);
  param.dim_X = size(X,1);
  param.dim_Y = size(Y,1);
  param.dim_U = 1;
  param.dim_V = 1;
  param.trial = 1;
  param.iter = 200;
  param.sigma_X = 1;
  param.sigma_Y = 1;
  param.verbose = 0;
  param.update_x = 1;
  param.update_y = 1;
  param.polish_step = 1;
  param.epsilon = 1e-5;

  param = setopt(param, varargin{:})
  param_optimizer.epsilon = param.epsilon;
  param_optimizer.line_search = 'naive'; %armijo';

  % --------------------------------------------------
  Xw = X / mean(sqrt(var(X, 1, 2)));
  Yw = Y / mean(sqrt(var(Y, 1, 2)));
  if 1
    Dx = distance(Xw,Xw);
    Dy = distance(Yw,Yw);
    if 1
      param.sigma_X = sqrt(median(Dx(:)));
      param.sigma_Y = sqrt(median(Dy(:))); 
    else
      param.sigma_X = sqrt(median(Dx(logical(triu(ones(size(Dx)))))));
      param.sigma_Y = sqrt(median(Dy(logical(triu(ones(size(Dy)))))));
    end
  end
  % [Xw Tx] = whitening(X);
  % [Yw Ty] = whitening(Y);

  for i = 1:param.dim_X
    XX{i} = ones(param.N,1)*Xw(i,:) - Xw(i,:)'*ones(1,param.N);
  end
  for i = 1:param.dim_Y
    YY{i} = ones(param.N,1)*Yw(i,:) - Yw(i,:)'*ones(1,param.N);
  end

  opt_obj = - inf;
  for t = 1:param.trial

    fprintf('\n%d/%d-th trial starts \n', t, param.trial);
    U = rand_mat(param.dim_X, param.dim_U);
    V = rand_mat(param.dim_Y, param.dim_V);
    
    % --------------------------------------------------
    % Main loop
    % --------------------------------------------------
    for iter = 1:param.iter
        
      if param.verbose
        fprintf('\n-- %d-th iteration -- \n', iter);
        U, V
      end
      UX = U * Xw;
      VY = V * Yw;

      [HSIC G] = calcHSIC(UX, VY, param, [], 'none');
      if param.verbose
        fprintf('HSIC = %f\n', HSIC);
      elseif mod(iter,10) == 1
        fprintf('\n-- %d-th iteration -- \n', iter);
        fprintf('HSIC = %f\n', HSIC);
      end
      rslt(t).log_HSIC(iter) = HSIC;
      
      [dU dV] = calc_deriv(UX, VY, XX, YY, G, param);

      [is_end U V param_optimizer] = ...
          ng_update(@calcHSIC, HSIC, Xw, Yw, U, V, dU, dV, [], param, param_optimizer);

      if is_end
        if strcmp(param_optimizer.line_search, 'naive')
          param_optimizer.line_search = 'armijo';
        else
          if param.polish_step == 1
            param.sigma_X = param.sigma_X / 2;
            param.sigma_Y = param.sigma_Y / 2;
            param.polish_step = 0;
          else
            break;
          end
        end
      end

    if opt_obj < HSIC
      opt_obj = HSIC;
      param.opt_U = U; %* Tx;
      param.opt_V = V; %* Ty;
      % param.U = param.U ./ repmat(sqrt(sum(param.U .* param.U, 2)),1,param.dim_X);
      % param.V = param.V ./ repmat(sqrt(sum(param.V .* param.V, 2)),1,param.dim_Y);
    end

    end % iteration
    
  end % trial

  % param.opt_U
  % param.opt_V


% --------------------------------------------------
function [HSIC G] = calcHSIC(UX, VY, param, model, cv)

  Du = distance(UX, UX);
  Dv = distance(VY, VY);

  G.Ku = exp(- Du ./ (2 * param.sigma_X^2));
  G.Kv = exp(- Dv ./ (2 * param.sigma_Y^2));
      
  G.Kuc = centering(G.Ku);
  G.Kvc = centering(G.Kv);

  HSIC = sum(sum(G.Kuc .* G.Kvc));


% --------------------------------------------------
function [dU dV] = calc_deriv(UX, VY, XX, YY, G, param)

  for i = 1:param.dim_U
    UU{i} = ones(param.N,1)*UX(i,:) - UX(i,:)'*ones(1,param.N);
  end
  for i = 1:param.dim_V
    VV{i} = ones(param.N,1)*VY(i,:) - VY(i,:)'*ones(1,param.N);
  end

  for k = 1:param.dim_U
    KuUU = G.Ku .* UU{k} ./ (param.sigma_X^2);
    for j = 1:param.dim_X
      dKu = - KuUU .* XX{j};
      dKuc = centering(dKu);
      dU(k,j) = sum(sum(dKuc .* G.Kvc));
    end
  end

  for k = 1:param.dim_V
    KvVV = G.Kv .* VV{k} ./ (param.sigma_Y^2);
    for j = 1:param.dim_Y
      dKv = - KvVV .* YY{j};
      dKvc = centering(dKv);
      dV(k,j) = sum(sum(G.Kuc .* dKvc));
    end
  end



% --------------------------------------------------
function [Kc] = centering(K)

  mK = mean(K);
  mK1 = ones(length(mK),1) * mK;
  Kc = K - mK1 - mK1' + repmat(mean(mK),size(K));

% --------------------------------------------------
function [Dw T] = whitening(D)

  covD = cov(D');
  T = inv(sqrtm(covD));
  Dw = T * (D - repmat(mean(D,2),1,size(D,2)));

% --------------------------------------------------
function M = rand_mat(d1, d2)
  M = randn(d1, d1); 
  M = M' * M;
  [M V D] = svd(M);
  M = M(1:d2, :);
  