function [param rslt] = ICCA(X, Y, varargin)

% ICCA - Informational Canonical Correlation Analysis
%
% Options
% 
% 'dim_U'         Projected dimension of X
% 'dim_V'         Projected dimension of Y
% 'sigma_list'    Candidates of KDE bandwidth
% 'trial'         The number of trials
% 'optimization'  Optimization strategy
% 'iter'          The maximum number of iteration
% 'U0'            Initial U
% 'V0'            Initial V

  % --------------------------------------------------
  param.dim_U = 1;
  param.dim_V = 1;
  % param.sigma_list = [];
  param.sigma_list = [0.1:0.025:0.9];
  param.trial = 10;
  param.opm = 'multi'; % one
  param.line_search = 'naive'; % armijo
  param.iter = 200;
  param.update_x = 1;
  param.update_y = 1;
  param.U0 = [];
  param.V0 = [];
  param.opt_U = [];
  param.opt_V = [];
  param.verbose = 0;
  update_UV = @sqp_update;
  param_optimizer.epsilon = 1e-5;
  tol = 1;

  param = setopt(param, varargin{:})  

  if strcmp(param.opm, 'one') && param.dim_U ~= param.dim_V
    error('One-dimensional optimization is only for dim_U == dim_V');
  end

  if strcmp(param.opm, 'one')
    outer_loop = param.dim_U;
    param.line_search = 'armijo'; 
  else
    outer_loop = 1;
    update_UV = @ng_update;
  end

  % --------------------------------------------------

  param.N = size(Y, 2);
  param.dim_Y = size(Y, 1);
  param.dim_X = size(X, 1);
  Y = Y / mean(sqrt(var(Y, 1, 2)));
  X = X / mean(sqrt(var(X, 1, 2)));

  for i = 1:param.dim_X
    XX{i} = ones(param.N,1)*X(i,:) - X(i,:)'*ones(1,param.N);
  end
  for i = 1:param.dim_Y
    YY{i} = ones(param.N,1)*Y(i,:) - Y(i,:)'*ones(1,param.N);
  end

  for out_dim = 1:outer_loop

    fprintf('%d-th outer loop \n', out_dim);

    opt_MI = -inf;
    for t = 1:param.trial

      fprintf('%d/%d-th trial starts \n', t, param.trial);
      [U V] = initialize(out_dim, param);


      update_sigma = 0; % update density estimation
      stagnate = 0; % for termination
      sigma = [];   % for KDE
      model = [];
      max_MI = -inf;
      param_optimizer.feasible = 1; % parameters for optimizer
      if strcmp(param.opm, 'multi')
        param_optimizer.line_search = 'naive';
      else
        param_optimizer.line_search = 'armijo';
      end
      % --------------------------------------------------
      % Main loop
      % --------------------------------------------------
      for iter = 1:param.iter
        

        if param.verbose
          fprintf('\n-- %d-th iteration -- \n', iter);
          U
          V
        end
        UX = U * X;
        VY = V * Y;
        if iter == 1 || iter == 5 || iter == 10 || update_sigma
          cv = 'update';
        else
          cv = 'none';
        end

        % --------------------------------------------------
        % Estimate MI
        % --------------------------------------------------
        [MI model P Ku Kv KuKv] = calcMI(UX, VY, param, model, cv);
        if update_sigma && strcmp(param.opm, 'one')
          param_optimizer.refresh_hessian = 1;
        end
        update_sigma = 0;

        if param.verbose
          fprintf('MI = %f\n', MI); 
        elseif  mod(iter,10) == 1
          fprintf('\n-- %d-th iteration -- \n', iter);
          fprintf('MI = %f\n', MI);
        end
        rslt(out_dim,t).log_MI(iter) = MI;       
        if max_MI < MI && param_optimizer.feasible 
          max_MI = MI;
          rslt(out_dim,t).MI = MI; rslt(out_dim,t).U = U; rslt(out_dim,t).V = V; rslt(out_dim,t).model = model;
        end

        % --------------------------------------------------
        % Update
        % --------------------------------------------------
        [dU dV] = calc_deriv(P, UX, VY, XX, YY, Ku, Kv, KuKv, model, param);

        param_optimizer.iter = iter;
        [is_end U V param_optimizer] = ...
            update_UV(@calcMI, MI, X, Y, U, V, dU, dV, model, param, param_optimizer);
        
        % --------------------------------------------------
        % Termination
        % --------------------------------------------------
        if is_end 
          if strcmp(param_optimizer.line_search, 'naive')%strcmp(param.line_search, 'naive')
            param_optimizer.line_search = 'armijo';
            % param.line_search = 'armijo';
          else
            stagnate = stagnate + 1;
          end
          update_sigma = 1;
        else
          stagnate = 0;
        end
        if stagnate > tol
          break;
        end

      end % iter      

      if opt_MI < rslt(out_dim,t).MI
        opt_MI = rslt(out_dim,t).MI;
        opt_rslt(out_dim) = rslt(out_dim,t);
      end
      
    end % trial

    opt_rslt
    %opt_rslt.U
    % opt_rslt.V
    
    if out_dim == 1
      param.max_MI(out_dim) = opt_rslt(out_dim).MI;
      param.opt_U = opt_rslt(out_dim).U;
      param.opt_V = opt_rslt(out_dim).V;
    else
      param.max_MI(out_dim) = opt_rslt(out_dim).MI;
      param.opt_U = [param.opt_U; opt_rslt(out_dim).U];
      param.opt_V = [param.opt_V; opt_rslt(out_dim).V];
    end

  end % outer_loop

% --------------------------------------------------
function [MI model P Ku Kv KuKv] = calcMI(UX, VY, param, model, cv)

  dim_U = size(UX, 1);
  dim_V = size(VY, 1);

  if strcmp(cv, 'update')
    if length(param.sigma_list) > 0 
      fprintf('Searching the optimal KDE bandwidth by LOO-CV ... \n');
      % -- CV --
      Du = distance(UX, UX);
      sigma_temp = CV_KDE(Du, param, dim_U, 1);
      model.sigma_U1 = sigma_temp * ones(1, dim_U);

      Dv = distance(VY, VY);
      sigma_temp = CV_KDE(Dv, param, dim_V, 1);
      model.sigma_V1 = sigma_temp * ones(1, dim_V);

      Dtemp{1} = Du;
      Dtemp{2} = Dv;
      sigma_temp = CV_KDE(Dtemp, param, [size(UX,1) size(VY,1)], 2);
      model.sigma_U2 = sigma_temp(1) * ones(1, dim_U);
      model.sigma_V2 = sigma_temp(2) * ones(1, dim_V);
    else
      fprintf('Setting KDE bandwidth by Silverman heuristics ... \n');
      % -- Silverman --
      s_UX = std(UX, 0, 2)';
      s_VY = std(VY, 0, 2)';

      if dim_U == 1 && dim_V == 1
        model.sigma_U1 = 1.06*s_UX*param.N^(-0.2);
        model.sigma_V1 = 1.06*s_VY*param.N^(-0.2);
      
        model.sigma_U2 = s_UX*param.N^(-1/6);
        model.sigma_V2 = s_VY*param.N^(-1/6);
      else
        model.sigma_U1 = s_UX.*param.N.^(-1/(dim_U+4));
        model.sigma_V1 = s_VY.*param.N.^(-1/(dim_V+4));

        model.sigma_U2 = s_UX.*param.N.^(-1/(dim_U+dim_V+4));
        model.sigma_V2 = s_VY.*param.N.^(-1/(dim_U+dim_V+4));
      end
      
    end
  end

  sigma_UX = diag(1./model.sigma_U1) * UX;
  sigma_VY = diag(1./model.sigma_V1) * VY;
  Du = distance(sigma_UX, sigma_UX);
  Dv = distance(sigma_VY, sigma_VY);

  Ku = exp(-Du ./ 2);
  Kv = exp(-Dv ./ 2);

  sigma2_UX = diag(1./model.sigma_U2) * UX;
  sigma2_VY = diag(1./model.sigma_V2) * VY;
  Du2 = distance(sigma2_UX, sigma2_UX);
  Dv2 = distance(sigma2_VY, sigma2_VY);

  KuKv = exp(-Du2 ./ 2) .* exp(-Dv2 ./ 2);
  
  P.U = mean(Ku) ./ ((2*pi)^(dim_U/2) * prod(model.sigma_U1));
  P.V = mean(Kv) ./ ((2*pi)^(dim_V/2) * prod(model.sigma_V1));
  P.UV = mean(KuKv) ./ ...
         ((2*pi)^(size(UX,1)/2) * prod(model.sigma_U2) * ...
          (2*pi)^(size(VY,1)/2) * prod(model.sigma_V2));

  MI = mean(log(P.UV) - log(P.U) - log(P.V));

% --------------------------------------------------
function [opt_sigma] = CV_KDE(D, param, dim, num_width)

  if num_width == 1
    max_logl = -inf;
    for sigma = param.sigma_list
      K = exp(-D ./ (2*(sigma^2)));
      p_LOO = (sum(K) - 1) ./ ...
              ((param.N - 1) * (2*pi*(sigma^2))^(dim/2));
      p_LOO(p_LOO < 10^(-14)) = 10^(-14);
      logl_cv = sum(log(p_LOO));
      if logl_cv > max_logl
        max_logl = logl_cv;
        opt_sigma = sigma;
      end
    end
  elseif num_width == 2
    max_logl = -inf;

    s_idx = 1;
    for sigma1 = param.sigma_list
      K1{s_idx} = exp(-D{1} ./ (2*(sigma1^2)));
      s_idx = s_idx + 1;
    end
    s2_idx = 1;
    for sigma2 = param.sigma_list
      K2{s2_idx} = exp(-D{2} ./ (2*(sigma2^2)));
      s2_idx = s2_idx + 1;
    end

    s_idx = 1;
    for sigma1 = param.sigma_list
      % K1 = exp(-D{1} ./ (2*(sigma1^2)));
      s2_idx = 1;
      for sigma2 = param.sigma_list
        % K2 = exp(-D{2} ./ (2*(sigma2^2)));
        K = K1{s_idx} .* K2{s2_idx};
        p_LOO = (sum(K) - 1) ./ ...
                ((param.N - 1) * ...
                 (2 * pi * (sigma1^2))^(dim(1)/2) * ...
                 (2 * pi * (sigma2^2))^(dim(2)/2));
        p_LOO(p_LOO < 10^(-14)) = 10^(-14);
        logl_cv = sum(log(p_LOO));
        if logl_cv > max_logl
          max_logl = logl_cv;
          opt_sigma(1) = sigma1;    
          opt_sigma(2) = sigma2;        
        end
        s2_idx = s2_idx + 1;
      end        
      s_idx = s_idx + 1;
    end

  else
    error('')
  end

% --------------------------------------------------
function [dU dV] = calc_deriv(P, UX, VY, XX, YY, Ku, Kv, KuKv, model, param)

  dim_U = size(UX, 1);
  dim_V = size(VY, 1);

  for i = 1:dim_U
    UU{i} = ones(param.N,1)*UX(i,:) - UX(i,:)'*ones(1,param.N);
  end
  for i = 1:dim_V
    VV{i} = ones(param.N,1)*VY(i,:) - VY(i,:)'*ones(1,param.N);
  end

  for k = 1:dim_U
    KuUU = Ku .* UU{k} ./ (model.sigma_U1(k)^2);
    KuKvUU = KuKv .* UU{k} ./ (model.sigma_U2(k)^2);
    for j = 1:param.dim_X
      dKu = - KuUU .* XX{j};
      dPu = mean(dKu) ./ ...
            ((2*pi)^(dim_U/2) * prod(model.sigma_U1));
        
      dKuKv = - KuKvUU .* XX{j}; 
      dPuv = mean(dKuKv) ./ ...
         ((2*pi)^(size(UX,1)/2) * prod(model.sigma_U2) * ...
          (2*pi)^(size(VY,1)/2) * prod(model.sigma_V2));

      temp = dPuv ./ P.UV - dPu ./ P.U;
      dU(k,j) = mean(temp);
    end
  end

  for k = 1:dim_V
    KvVV = Kv .* VV{k} ./ (model.sigma_V1(k)^2);
    KuKvVV = KuKv .* VV{k} ./ (model.sigma_V2(k)^2);
    for j = 1:param.dim_Y
      dKv = - KvVV .* YY{j};
      dPv = mean(dKv) ./ ...
            ((2*pi)^(dim_V/2) * prod(model.sigma_V1));
        
      dKuKv = - KuKvVV .* YY{j};
      dPuv = mean(dKuKv) ./ ...
         ((2*pi)^(size(UX,1)/2) * prod(model.sigma_U2) * ...
          (2*pi)^(size(VY,1)/2) * prod(model.sigma_V2));
      temp = dPuv ./ P.UV - dPv ./ P.V;
      dV(k,j) = mean(temp);
        
    end
  end

  if param.verbose
    dU
    dV
  end
