function [param rslt] = KCCA_CDA(X, Y, varargin)

% KCCA_CDA - CDA by maximizing F-correlation or Kernel Generalized Variance (KGV)
%
% Options
% 'dim_U'         Dimension of projection matrix for X      1
% 'dim_V'         Dimension of projection matrix for Y      1
% 'sigma'         Gaussian kernel parameters           
% 'kappa'         Regularization parameters           
% 'trial'         The number of trials                      10
% 'contrast'      Contrast function ('kcca'|'kgv')          kgv
% 'iter'          The maximum number of iteration           200

% --------------------------------------------------

  param.N = size(X,2);
  param.dim_X = size(X,1);
  param.dim_Y = size(Y,1);
  param.dim_U = 1;
  param.dim_V = 1;
  param.trial = 1;
  % param.contrast = 'kcca';
  param.contrast = 'kgv';
  param.iter = 200;
  if param.N > 1000 % F. Bach's heuristics
    param.sigma = 0.5;
    param.kappa = 0.002 * param.N / 2;
  else
    param.sigma = 1;
    param.kappa = 0.02 * param.N / 2;
  end
  param.verbose = 0;
  param.update_x = 1;
  param.update_y = 1;
  param.epsilon = 1e-5;

  param = setopt(param, varargin{:}) 
  param_optimizer.epsilon = param.epsilon;
  param_optimizer.line_search = 'naive'; %armijo';

  % --------------------------------------------------
  % Xw = X / mean(sqrt(var(X, 1, 2)));
  % Yw = Y / mean(sqrt(var(Y, 1, 2)));
  [Xw Tx] = whitening(X);
  [Yw Ty] = whitening(Y);

  Tx
  Ty

  for i = 1:param.dim_X
    XX{i} = ones(param.N,1)*Xw(i,:) - Xw(i,:)'*ones(1,param.N);
  end
  for i = 1:param.dim_Y
    YY{i} = ones(param.N,1)*Yw(i,:) - Yw(i,:)'*ones(1,param.N);
  end

  opt_obj = - inf;
  for t = 1:param.trial

    fprintf('%d/%d-th trial starts \n', t, param.trial);
    U = rand_mat(param.dim_X, param.dim_U);
    V = rand_mat(param.dim_Y, param.dim_V);
    
    % --------------------------------------------------
    % Main loop
    % --------------------------------------------------
    for iter = 1:param.iter
        
      if param.verbose
        fprintf('\n-- %d-th iteration -- \n', iter);
        U, V
      end
      UX = U * Xw;
      VY = V * Yw;

      [I G] = calcContrast(UX, VY, param, [], 'none');

      if param.verbose
        fprintf('I = %f\n', I);
      elseif mod(iter,10) == 1
        fprintf('\n-- %d-th iteration -- \n', iter);
        fprintf('I = %f\n', I);
      end
      rslt(t).log_I(iter) = I;

      [dU dV] = calc_deriv(UX, VY, XX, YY, G, param);

      [is_end U V param_optimizer] = ...
          ng_update(@calcContrast, I, Xw, Yw, U, V, dU, dV, [], param, param_optimizer);

      if is_end
        if strcmp(param_optimizer.line_search, 'naive')
          param_optimizer.line_search = 'armijo'
        else
          break;
        end
      end

    end % iteration
    
    rslt(t).U = U;
    rslt(t).V = V;

    if opt_obj < I
      opt_obj = I;
      U = U * Tx; V = V * Ty;
      U = sqrt(diag(1./sum(U' .* U'))) * U;   
      V = sqrt(diag(1./sum(V' .* V'))) * V;
      param.opt_U = U; param.opt_V = V;
      % param.opt_U = U * Tx; param.opt_V = V * Ty;
    end

  end % trial

  % param.opt_U
  % param.opt_V

% --------------------------------------------------
function [I G] = calcContrast(UX, VY, param, model, cv)

  Du = distance(UX, UX);
  Dv = distance(VY, VY);

  G.Ku = exp(- Du ./ (2 * param.sigma^2));
  G.Kv = exp(- Dv ./ (2 * param.sigma^2));
      
  G.Kuc = centering(G.Ku);
  G.Kvc = centering(G.Kv);
      
  % If we want to reduce computational cost,
  % you can use low-rank approximation of Gram matrix.
  kappaI = param.kappa * eye(param.N);
  G.invKucI = inv(G.Kuc + kappaI);
  G.invKvcI = inv(G.Kvc + kappaI);

  G.rKu = G.invKucI * G.Kuc;
  G.rKv = G.invKvcI * G.Kvc;
  
  G.rKurKv = G.rKu * G.rKv;

  G.Ktilde = [eye(param.N) G.rKurKv; ...
              G.rKurKv' eye(param.N)];
  
  switch param.contrast
   case 'kcca'
    OPTIONS.disp=0;
    OPTIONS.tol=1e-5;
    [beta lambda] = eigs(G.Ktilde,1,'SM',OPTIONS);
    G.beta = beta;
    G.lambda = lambda;
    I = - log(lambda) / 2;
   case 'kgv'
    I =  -log(det(G.Ktilde)) / 2;
  end


% --------------------------------------------------
function [dU dV] = calc_deriv(UX, VY, XX, YY, G, param)

  for i = 1:param.dim_U
    UU{i} = ones(param.N,1)*UX(i,:) - UX(i,:)'*ones(1,param.N);
  end
  for i = 1:param.dim_V
    VV{i} = ones(param.N,1)*VY(i,:) - VY(i,:)'*ones(1,param.N);
  end

  switch param.contrast
   case 'kcca'

    for k = 1:param.dim_U
      KuUU = G.Ku .* UU{k} ./ (param.sigma^2);
      for j = 1:param.dim_X
        dKu = - KuUU .* XX{j};
        dKuc = centering(dKu);
        drKu = G.invKucI * dKuc * (eye(param.N) - G.rKu);
        drKurKv = drKu * G.rKv;
        dU(k,j) = - G.beta' * ([zeros(param.N) drKurKv; drKurKv' zeros(param.N)] ...
                             * G.beta) / (2 * G.lambda);
      end
    end

    for k = 1:param.dim_V
      KvVV = G.Kv .* VV{k} ./ (param.sigma^2);
      for j = 1:param.dim_Y
        dKv = - KvVV .* YY{j};
        dKvc = centering(dKv);
        drKv = G.invKvcI * dKvc * (eye(param.N) - G.rKv);
        drKurKv = G.rKu * drKv;
        dV(k,j) = - G.beta' * ([zeros(param.N) drKurKv; drKurKv' zeros(param.N)] ...
                             * G.beta) / (2 * G.lambda);
      end
    end

   case 'kgv'

    invKtilde = inv(G.Ktilde);   

    for k = 1:param.dim_U
      KuUU = G.Ku .* UU{k} ./ (param.sigma^2);
      % KuUU = G.Kuc .* UU{k} ./ (param.sigma^2);
      for j = 1:param.dim_X
        dKu = - KuUU .* XX{j};
        dKuc = centering(dKu);
        drKu = G.invKucI * dKuc * (eye(param.N) - G.rKu);
        drKurKv = drKu * G.rKv;
        dU(k,j) = -(sum(sum(invKtilde(1:param.N,param.N+1:end) .* (drKurKv'),2)) + ...
                     sum(sum(invKtilde(param.N+1:end,1:param.N) .* drKurKv,2))) / 2;
      end
    end

    for k = 1:param.dim_V
      % KvVV = G.Kvc .* VV{k} ./ (param.sigma^2);
      KvVV = G.Kv .* VV{k} ./ (param.sigma^2);
      for j = 1:param.dim_Y
        dKv = - KvVV .* YY{j};
        dKvc = centering(dKv);
        drKv = G.invKvcI * dKvc * (eye(param.N) - G.rKv);
        drKurKv = G.rKu * drKv;
        dV(k,j) = -(sum(sum(invKtilde(1:param.N,param.N+1:end) .* (drKurKv'),2)) + ...
                     sum(sum(invKtilde(param.N+1:end,1:param.N) .* drKurKv,2))) / 2;
      end
    end
    
  end


% --------------------------------------------------
function [Kc] = centering(K)

mK = mean(K);

mK1 = ones(length(mK),1) * mK;

Kc = K - mK1 - mK1' + repmat(mean(mK),size(K));

% --------------------------------------------------
function [Dw T] = whitening(D)

  covD = cov(D');
  T = inv(sqrtm(covD));
  Dw = T * (D - repmat(mean(D,2),1,size(D,2)));

% --------------------------------------------------
function M = rand_mat(d1, d2)
  M = randn(d1, d1); 
  M = M' * M;
  [M V D] = svd(M);
  M = M(1:d2, :);
  