function [param log_LSCDA] = LSCDA(X, Y, varargin)

% LSCDA - Least Squares Canonical Dependency Analysis
%
% Options                                                   Default
%
% 'dim_U'         Dimension of projection matrix for X      1
% 'dim_V'         Dimension of projection matrix for Y      1
% 'sigma_list'    Candidates of kernel parameters           [0.1:0.025:0.9]
% 'lambda_list'   Candidates of regularization paramter     logspace(-3:-1:7)
% 'trial'         The number of trials                      10
% 'opm'           Optimization strategy ('multi'|'one')     'multi'
% 'iter'          The maximum number of iteration           500
% 'U0'            Initial U                                 none
% 'V0'            Initial V                                 none
% 'b'             The number of basis                       200
% 'fold_num'      'fold' of CV                              10

  % --------------------------------------------------
  param.dim_U = 1;
  param.dim_V = 1;
  param.obj = 0;
  param.sigma_list = [0.1:0.025:0.9];
  param.lambda_list = logspace(-3,-1,7);
  param.trial = 10;
  param.opm = 'multi'; % one
  % param.opm = 'one'; 
  % param.line_search = 'naive'; % armijo
  param.iter = 500;
  param.b = 500;
  param.fold_num = 5; % 5;
  param.update_x = 1;
  param.update_y = 1;
  param.U0 = [];
  param.V0 = [];
  param.opt_U = [];
  param.opt_V = [];
  param.verbose = 0;
  param.epsilon = 1e-5; %1e-5;
  update_UV = @ng_update;

  coef_I = 1; % 0.0001;
  coef_G = 0;
  tol = 1;

  param = setopt(param, varargin{:})
  param_optimizer.termination = 'abs'; % abs | relative
  param_optimizer.epsilon = param.epsilon; 

  if strcmp(param.opm, 'one') && param.dim_U ~= param.dim_V
    error('One-dimensional optimization is only for dim_U == dim_V');
  end

  if strcmp(param.opm, 'one')
    outer_loop = param.dim_U;
    update_UV = @sqp_update;
  else
    outer_loop = 1;
  end

  % --------------------------------------------------

  param.N = size(Y, 2);
  param.dim_Y = size(Y, 1);
  param.dim_X = size(X, 1);
  param.b = min(param.N, param.b);
  random_idx = randperm(param.N);
  param.ce_idx = random_idx(1:param.b); 
  param.cv_idx = randperm(param.N);
  param.cv_split = floor([0:param.N-1]*param.fold_num./param.N) + 1;

  Y = Y / mean(sqrt(var(Y, 1, 2)));
  X = X / mean(sqrt(var(X, 1, 2)));
  if strcmp(param.opm, 'multi')
    Y = Y / sqrt(param.dim_V);
    X = X / sqrt(param.dim_U);
  end
 
  for i = 1:param.dim_X
    XX{i} = ones(param.b,1)*X(i,:) - X(i,param.ce_idx)'*ones(1,param.N);
  end
  for i = 1:param.dim_Y
    YY{i} = ones(param.b,1)*Y(i,:) - Y(i,param.ce_idx)'*ones(1,param.N);
  end

  for out_dim = 1:outer_loop

    if outer_loop > 1
      fprintf('%d/%d-th outer loop \n', out_dim, outer_loop);
    end

    for t = 1:param.trial

      fprintf('%d/%d-th trial starts \n', t, param.trial);

      param.cv_idx = randperm(param.N);
      param.cv_split = floor([0:param.N-1]*param.fold_num ./ param.N) + 1;
      param.ce_idx = random_idx(1:param.b); 
      [U V] = initialize(out_dim, param);

      update_sigma = 0; % update density estimation
      stagnate = 0; % for termination
      max_SMI = -inf;

      model = [];
      model.coef_I = coef_I; 
      model.coef_G = coef_G; 
      param_optimizer.feasible = 1;
      update_model = 1;
      cv_method = 'sub2';
      if strcmp(param.opm, 'multi')
        param_optimizer.line_search = 'naive';
      else
        param_optimizer.line_search = 'armijo';
      end
      % --------------------------------------------------
      % Main loop
      % --------------------------------------------------
      for iter = 1:param.iter
        if param.verbose
          fprintf('\n-- %d-th iteration -- \n', iter);
          U
          V 
        end
        UX = U * X;
        VY = V * Y;

        % --------------------------------------------------
        % Estimate SMI
        % --------------------------------------------------
        if update_model || iter == 1 || iter == 5 || iter == 10 
          cv = cv_method;
        else
          cv = 'none';
        end
        [SMI model Hhat hhat Ku Kv KuKv] = calcSMI(UX, VY, param, model, cv);
        update_model = 0;

        if param.verbose
          fprintf('SMI = %f\n', SMI);
        elseif mod(iter,10) == 1
          fprintf('\n-- %d-th iteration -- \n', iter);
          fprintf('SMI = %f\n', SMI);
        end
        log_LSCDA(out_dim,t).SMI(iter) = SMI;
        if max_SMI < SMI && param_optimizer.feasible 
          max_SMI = SMI;
          log_LSCDA(out_dim,t).max_SMI = SMI; 
          log_LSCDA(out_dim,t).U = U; log_LSCDA(out_dim,t).V = V; 
          log_LSCDA(out_dim,t).model = model;
        end

        % --------------------------------------------------
        % Update U, V
        % --------------------------------------------------
        [dU dV] = calc_deriv(UX, VY, XX, YY, Hhat, hhat, Ku, Kv, KuKv, model, param);

        param_optimizer.iter = iter;
        [is_end U V param_optimizer model] = ...
            update_UV(@calcSMI, SMI, X, Y, U, V, dU, dV, model, param, param_optimizer);

        % --------------------------------------------------
        % Termination
        % --------------------------------------------------
        if is_end 

          if strcmp(cv_method, 'sub2') 
            cv_method = 'ps3';
            % cv_method = 'ps2';
          end

          if strcmp(param_optimizer.line_search, 'naive')
            param_optimizer.line_search = 'armijo';
          else
            stagnate = stagnate + 1;
          end
          update_model = 1;
        else
          stagnate = 0;
        end
        if stagnate > tol
          break;
        end

      end % iter

      log_LSCDA(out_dim,t).SMIcv = cv_LSMI(X, Y, log_LSCDA(out_dim,t).U, log_LSCDA(out_dim,t).V, log_LSCDA(out_dim,t).model, param);
      
      log_LSCDA(out_dim,t)
      % log_LSCDA(out_dim,t).U
      % log_LSCDA(out_dim,t).V

    end % trial

    [param.max_SMIcv(out_dim) idx] = max([log_LSCDA(out_dim,:).SMIcv]);
    param.SMI(out_dim) = log_LSCDA(out_dim,t).max_SMI;
    if out_dim == 1
      param.opt_U = log_LSCDA(out_dim,idx).U;
      param.opt_V = log_LSCDA(out_dim,idx).V;
    else
      param.opt_U = [param.opt_U; log_LSCDA(out_dim,idx).U];
      param.opt_V = [param.opt_V; log_LSCDA(out_dim,idx).V];
    end

  end % outer_loop


% --------------------------------------------------
function [SMI model Hhat hhat Ku Kv KuKv] = calcSMI(UX, VY, param, model, cv)

  dim_U = size(UX, 1);
  dim_V = size(VY, 1);

  Du = distance(UX(:,param.ce_idx), UX);
  Dv = distance(VY(:,param.ce_idx), VY);

  if ~strcmp(cv, 'none')
    if param.verbose
      fprintf('Searching the optimal density-ratio model by CV ... \n');
    end
    model = CV_LSCDA(Du, Dv, param, model, cv);
  end

  Ku = exp(- Du ./ (2*model.sigmaX^2));
  Kv = exp(- Dv ./ (2*model.sigmaY^2));  
  KuKv = Ku .* Kv;
  model.RegMat = model.coef_G*KuKv(:,param.ce_idx) + model.coef_I*eye(param.b);
  
  Hhat = (Ku * Ku') .* (Kv * Kv') ./ (param.N^2);
  hhat = mean(KuKv, 2);
  model.alpha = (Hhat + model.lambda*model.RegMat) \ hhat;
  % beta = (Hhat + model.lambda*RegMat) \ (Hhat * alpha);
  
  % alpha
  % ratio = alpha' * KuKv;
  SMI = calc_Is(Hhat, hhat, model.alpha, param.obj); 

% --------------------------------------------------
function Is = calc_Is(Hhat, hhat, alpha, obj)
  if obj == 0
    Is = - alpha'*Hhat*alpha/2 + hhat'*alpha - 1/2;
  else
    Is = hhat'*alpha/2 - 1/2;
  end

% --------------------------------------------------
function [dU dV] = calc_deriv(UX, VY, XX, YY, Hhat, hhat, Ku, Kv, KuKv, model, param)
  %obj = 0;

  dim_U = size(UX, 1);
  dim_V = size(VY, 1);

  for i = 1:dim_U
    UU{i} = ones(param.b,1)*UX(i,:) - UX(i,param.ce_idx)'*ones(1,param.N);
  end
  for i = 1:dim_V
    VV{i} = ones(param.b,1)*VY(i,:) - VY(i,param.ce_idx)'*ones(1,param.N);
  end

  beta = (Hhat + model.lambda*model.RegMat) \ (Hhat * model.alpha);

  Ku2 = Ku * Ku';
  Kv2 = Kv * Kv';
  Kv_v = repmat(Kv,param.dim_X,1);
  for k = 1:dim_U
    KuUU = Ku .* UU{k};
    % <--
%     dKu_temp = - (repmat(KuUU,param.dim_X,1) .* vertcat(XX{:})) ./ (model.sigmaX^2);
%     dKu2_temp = dKu_temp * Ku';
%     dhhat_temp = mat2cell(mean(dKu_temp .* Kv_v, 2),param.b*ones(1,param.dim_X),1);
%     dKu_temp2 = mat2cell(dKu_temp,param.b*ones(1,param.dim_X),param.N);
%     dKu2_temp2 = mat2cell(dKu2_temp,param.b*ones(1,param.dim_X),param.b);
    % -->
    for j = 1:param.dim_X
      dKu = - KuUU .* XX{j} ./ (model.sigmaX^2);
      dKu2 = dKu * Ku';
      % dKu = dKu_temp2{j};
      % dKu2 = dKu2_temp2{j};
      % dhhat = dhhat_temp{j};
      dhhat = mean(dKu .* Kv, 2);
      dHhat = (dKu2 + dKu2') .* Kv2 / (param.N^2);
      if model.coef_G > 0
        dR = model.coef_G .* dKu(:,param.ce_idx) .* Kv(:,param.ce_idx);
      else
        dR = zeros(param.b);
      end
      if param.obj == 0
        dU(k,j) = dhhat'*(2*model.alpha-beta) - model.alpha'*dHhat*(1.5*model.alpha-beta) + ...
                  model.lambda*model.alpha'*dR*(beta-model.alpha);
      else
        % alpha^\top h
        dU(k,j) = (- model.alpha'*(dHhat + model.lambda*dR)*model.alpha + 2 * model.alpha' * dhhat) / 2;
      end
    end
  end

  for k = 1:dim_V
    KvVV = Kv .* VV{k};
    for j = 1:param.dim_Y
      dKv = - KvVV .* YY{j} ./ (model.sigmaY^2);
      dKv2 = dKv * Kv';
      dhhat = mean(Ku .* dKv, 2);
      dHhat = Ku2 .* (dKv2 + dKv2') / (param.N^2);
      if model.coef_G > 0
        dR = model.coef_G .* Ku(:,param.ce_idx) .* dKv(:,param.ce_idx);
      else
        dR = zeros(param.b);
      end
      if param.obj == 0
        %% - 0.5 alpha^\top H \alpha + alpha^\top h
        dV(k,j) = dhhat'*(2*model.alpha-beta) - model.alpha'*dHhat*(1.5*model.alpha-beta) ...
                  + model.lambda*model.alpha'*dR*(beta-model.alpha);
      else
        %% alpha^\top h
        dV(k,j) = (- model.alpha'*(dHhat + model.lambda*dR) *model.alpha + 2 * model.alpha' * dhhat) / 2;
      end
    end
  end

  if param.verbose
    dU
    dV
  end


% --------------------------------------------------
function [SMIcv] = cv_LSMI(X, Y, U, V, model, param)
  % obj = 0;

  UX = U * X;
  VY = V * Y;
  Du = distance(UX(:,param.ce_idx), UX);
  Dv = distance(VY(:,param.ce_idx), VY);
  Ku = exp(- Du ./ (2*model.sigmaX^2));
  Kv = exp(- Dv ./ (2*model.sigmaY^2));
  KuKv = Ku .* Kv;
  RegMat = model.coef_G*KuKv(:,param.ce_idx) + model.coef_I*eye(param.b);

  % CV 

  for k = 1:param.fold_num 
    tr_idx = param.cv_idx(param.cv_split ~= k);
    va_idx = param.cv_idx(param.cv_split == k);
    tr_ce_idx = ismember(param.ce_idx, tr_idx); 
            
    %-- Train --
    alpha_cv = uLSIF(Ku(tr_ce_idx,tr_idx), Kv(tr_ce_idx,tr_idx), ...
                     model.lambda*(param.b/nnz(tr_ce_idx)), RegMat(tr_ce_idx,tr_ce_idx));
                     % model.lambda, RegMat(tr_ce_idx,tr_ce_idx));
                     % model.lambda*(nnz(tr_ce_idx)/param.b), RegMat(tr_ce_idx,tr_ce_idx));
    
    %-- Validation --
    n_va = length(va_idx);
    Ku_va = Ku(tr_ce_idx,va_idx);
    Kv_va = Kv(tr_ce_idx,va_idx);
    Hhat_va = (Ku_va * Ku_va') .* (Kv_va * Kv_va') ./ (n_va^2);        

    Is_cv(k) = calc_Is(Hhat_va, mean(Ku_va .* Kv_va, 2), alpha_cv, param.obj);

  end

  SMIcv = mean(Is_cv);

